<?php

/**
 * @file
 * Contains \Drupal\cdek_api\Cdek.
 */

namespace Drupal\cdek_api;

use Drupal\cdek_api\Exception\CalculatorApiException;
use Symfony\Component\Serializer\Encoder\JsonEncoder;
use Symfony\Component\Serializer\Encoder\XmlEncoder;
use Drupal\Core\Cache\CacheBackendInterface;
use Symfony\Component\Serializer\Serializer;
use Drupal\Component\Utility\Crypt;
use Drupal\Core\Cache\Cache;

/**
 * Wrapper class for CDEK API.
 *
 * See CDEK API documentation for more info.
 */
class Cdek {

  /**
   * The key for the "all" parameter.
   */
  const PARAMS_ALL = 'all';

  /**
   * CDEK API settings.
   *
   * @var array
   */
  protected $settings = [];

  /**
   * CDEK API request manager.
   *
   * @var \Drupal\cdek_api\RequestManager
   */
  protected $requestManager;

  /**
   * The cache object associated with the 'cdek_api' bin.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * The serializer.
   *
   * @var \Symfony\Component\Serializer\Serializer
   */
  protected $serializer;

  /**
   * Cdek constructor.
   *
   * @param \Drupal\cdek_api\RequestManager $request_manager
   *   CDEK API request manager.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache object associated with the 'cdek_api' bin.
   */
  public function __construct(RequestManager $request_manager, CacheBackendInterface $cache) {
    $this->requestManager = $request_manager;
    $this->cache = $cache;
    // Create an instance of the serializer.
    $encoders[] = new JsonEncoder();
    $encoders[] = new XmlEncoder();
    $this->serializer = new Serializer([], $encoders);
    // Get the configuration values.
    $this->settings = self::getSettings();
  }

  /**
   * Gets CDEK API settings.
   *
   * @return array
   *   Array with settings.
   */
  public static function getSettings() {
    $settings = \Drupal::config('cdek_api.settings')->get();
    unset($settings['_core']);
    return $settings;
  }

  /**
   * Load order pickup points.
   *
   * @param array $params
   *   API function params for filtered results.
   *   The following parameters are allowed:
   *   - cityid - City code in CDEK system.
   *   - citypostcode - City postcode string, cityid is more priority.
   *   - type - Pickup point type. Allowed values: "PVZ", "POSTOMAT", "ALL".
   *     Default value is "PVZ".
   *   - countryid - Country code in CDEK system.
   *   - regionid - Region code in CDEK system.
   *   - havecashless - (boolean) Is there a payment terminal.
   *   - isdressingroom - (boolean) Is there a fitting room in the pickup point.
   *   - weightmax - The maximum weight that can be sent to the pickup point.
   *   - allowedcod - (boolean) CASH is allowed in the pickup point.
   *
   * @return array|null
   *   Order pickup points array or NULL on failure.
   */
  public function getOrderPickupPoints(array $params = []) {
    static $pickup_points = [];
    $hash = $this->generateParamsHash($params);

    // Check static data.
    if (isset($pickup_points[$hash])) {
      return $pickup_points[$hash];
    }
    // Check cache data.
    $cid = "order_pickup_points:$hash";
    if ($this->isCached() && ($cache_data = $this->getCacheData($cid)) !== FALSE) {
      return $pickup_points[$hash] = $cache_data;
    }

    // Sending API request.
    $result = $this->requestManager->sendOrderPickupPointsRequest($params);
    if ($result === NULL) {
      return NULL;
    }
    try {
      // Convert XML to an associative array.
      $data = $this->serializer->decode($result, 'xml');
      // Check the presence of the element with order pickup points.
      if (!isset($data['Pvz']) || !is_array($data['Pvz'])) {
        throw new \UnexpectedValueException('Missing element with order pickup points.');
      }
    }
    catch (\UnexpectedValueException $e) {
      // Failed to decode XML.
      return NULL;
    }
    // Convert data to a more convenient form.
    $data = $this->clearXmlDecodingResult($data['Pvz']);
    if (isset($data['Code'])) {
      $data = [$data];
    }
    foreach ($data as $key => $value) {
      $data[$value['Code']] = $value;
      unset($data[$key]);
    }

    // Set persistent and static cache.
    $pickup_points[$hash] = $data;
    if ($this->isCached()) {
      $this->setCacheData($cid, $pickup_points[$hash]);
    }
    return $pickup_points[$hash];
  }

  /**
   * Load order pickup point.
   *
   * @param string $code
   *   The code of the point to load.
   *
   * @return array|null
   *   The point array or NULL if there is no point with the given code.
   */
  public function getOrderPickupPoint($code) {
    $pickup_points = $this->getOrderPickupPoints();
    return isset($pickup_points[$code]) ? $pickup_points[$code] : NULL;
  }

  /**
   * Load countries.
   *
   * @return array|null
   *   Country names array keyed by country code, or NULL on failure.
   */
  public function getCountries() {
    static $countries;

    // Check static data.
    if (isset($countries)) {
      return $countries;
    }
    // Check cache data.
    if ($this->isCached() && ($cache_data = $this->getCacheData('countries')) !== FALSE) {
      return $countries = $cache_data;
    }

    // Load order pickup points.
    $pickup_points = $this->getOrderPickupPoints();
    if ($pickup_points === NULL) {
      return NULL;
    }

    // Get countries from pickup points.
    $countries = [];
    foreach ($pickup_points as $pickup_point) {
      if (!isset($countries[$pickup_point['CountryCode']])) {
        $countries[$pickup_point['CountryCode']] = $pickup_point['CountryName'];
      }
    }
    // Sort countries in alphabet.
    asort($countries);
    // Set persistent cache.
    if ($this->isCached()) {
      $this->setCacheData('countries', $countries);
    }
    return $countries;
  }

  /**
   * Load regions by country.
   *
   * @param int $country
   *   Country code in CDEK system.
   *
   * @return array|null
   *   Region names array keyed by region code, or NULL on failure.
   */
  public function getRegions($country = NULL) {
    static $regions = [];

    // Prepare API request params.
    $params = ['countryid' => $country];
    if ($country === NULL) {
      $country = self::PARAMS_ALL;
      unset($params['countryid']);
    }

    // Check static data.
    if (isset($regions[$country])) {
      return $regions[$country];
    }
    // Check cache data.
    $cid = "regions:$country";
    if ($this->isCached() && ($cache_data = $this->getCacheData($cid)) !== FALSE) {
      return $regions[$country] = $cache_data;
    }

    // Load order pickup points.
    $pickup_points = $this->getOrderPickupPoints($params);
    if ($pickup_points === NULL) {
      return NULL;
    }

    // Get regions from pickup points.
    $regions[$country] = [];
    foreach ($pickup_points as $pickup_point) {
      if (!isset($regions[$country][$pickup_point['RegionCode']])) {
        $regions[$country][$pickup_point['RegionCode']] = $pickup_point['RegionName'];
      }
    }
    // Sort regions in alphabet.
    asort($regions[$country]);
    // Set persistent cache.
    if ($this->isCached()) {
      $this->setCacheData($cid, $regions[$country]);
    }
    return $regions[$country];
  }

  /**
   * Load cities by country and/or region.
   *
   * @param int $country
   *   Country code in CDEK system.
   * @param int $region
   *   Region code in CDEK system.
   *
   * @return array|null
   *   City names array keyed by city code, or NULL on failure.
   */
  public function getCities($country = NULL, $region = NULL) {
    static $cities = [];

    // Prepare API request params.
    $params = ['countryid' => $country, 'regionid' => $region];
    if ($country === NULL) {
      $country = self::PARAMS_ALL;
      unset($params['countryid']);
    }
    if ($region === NULL) {
      $region = self::PARAMS_ALL;
      unset($params['regionid']);
    }

    // Check static data.
    if (isset($cities[$country][$region])) {
      return $cities[$country][$region];
    }
    // Check cache data.
    $cid = "cities:$country:$region";
    if ($this->isCached() && ($cache_data = $this->getCacheData($cid)) !== FALSE) {
      return $cities[$country][$region] = $cache_data;
    }

    // Load order pickup points.
    $pickup_points = $this->getOrderPickupPoints($params);
    if ($pickup_points === NULL) {
      return NULL;
    }

    // Get cities from pickup points.
    $cities[$country][$region] = [];
    foreach ($pickup_points as $pickup_point) {
      if (!isset($cities[$country][$region][$pickup_point['CityCode']])) {
        $cities[$country][$region][$pickup_point['CityCode']] = $pickup_point['City'];
      }
    }
    // Sort cities in alphabet.
    asort($cities[$country][$region]);
    // Set persistent cache.
    if ($this->isCached()) {
      $this->setCacheData($cid, $cities[$country][$region]);
    }
    return $cities[$country][$region];
  }

  /**
   * Calculate delivery rate by params.
   *
   * @param array $params
   *   API function params to calculate.
   *   The following parameters are allowed:
   *   - authLogin - Login for CDEK account. Set TRUE to use the login
   *     specified in the configuration.
   *   - secure - Password for CDEK account. If authLogin is set to TRUE,
   *     then the password will be taken from the configuration.
   *   - dateExecute - Planning date of shipment in the format YYYY-ММ-DD.
   *   - senderCityId - Origin city code in CDEK system.
   *   - senderCityPostCode - Zip code of the city of origin.
   *   - receiverCityId - Destination city code in CDEK system.
   *   - receiverCityPostCode - Zip code of the destination city.
   *   - tariffId - Code of chosen tariff.
   *   - modeId - Chosen delivery mode.
   *   - goods - Array with elements each of which contains an
   *     associative array with keys:
   *       - weight - float, piece weight, kg.
   *       - length - integer, piece length, sm.
   *       - width - integer, piece width, sm.
   *       - height - integer, depth, sm.
   *     Or with such kind of parameters:
   *       - weight - float, piece weight, kg.
   *       - volume - float, piece volume, cum.
   *
   * @return array|null
   *   An array with delivery rate information, or NULL on failure.
   *
   * @throws \Drupal\cdek_api\Exception\CalculatorApiException
   *   Delivery rate calculation API exception.
   */
  public function calculateDeliveryRate(array $params = []) {
    // TODO: Define a constant for "version" or move it to a configuration.
    // TODO: Add "tariffList" and "services" params support.
    // Merge to defaults.
    $params += [
      'version' => '1.0',
      'senderCityId' => $this->settings['sender_city_id'],
      'senderCityPostCode' => $this->settings['sender_city_postcode'],
      'tariffId' => $this->settings['tariff_id'],
    ];

    // Use authorization if necessary.
    if (isset($params['authLogin']) && $params['authLogin'] === TRUE) {
      $params['authLogin'] = $this->settings['auth_login'];
      $params['secure'] = $this->settings['secure_password'];
    }
    // Process "secure" param.
    if (isset($params['secure'])) {
      if (!isset($params['dateExecute'])) {
        $params['dateExecute'] = date('Y-m-d', time());
      }
      $params['secure'] = md5($params['dateExecute'] . '&' . $params['secure']);
    }
    // Filter empty values.
    $params = array_filter($params);

    // Sending API request.
    $result = $this->requestManager->sendDeliveryRateCalculationRequest($params);
    if ($result === NULL) {
      return NULL;
    }
    try {
      // Convert JSON to an associative array.
      $data = $this->serializer->decode($result, 'json');
      // Check API errors.
      if (isset($data['error'])) {
        // Throw exception by first error.
        $error = array_shift($data['error']);
        if (isset($error['text']) && isset($error['code'])) {
          throw new CalculatorApiException($error['text'], $error['code']);
        }
      }
      // Check the presence of the result element.
      if (!isset($data['result']) || !is_array($data['result'])) {
        throw new \UnexpectedValueException('Missing result element.');
      }
    }
    catch (\UnexpectedValueException $e) {
      // Failed to decode JSON.
      return NULL;
    }
    return $data['result'];
  }

  /**
   * Verifying that the data is cached.
   *
   * @return bool
   *   API data is cached (TRUE) or not (FALSE).
   */
  public function isCached() {
    return $this->settings['cache_expire'] !== NULL;
  }

  /**
   * Stores API data in the persistent cache.
   *
   * @param string $cid
   *   The cache ID of the data to store.
   * @param mixed $data
   *   The data to store in the cache.
   */
  protected function setCacheData($cid, $data) {
    $expire = $this->settings['cache_expire'];
    if ($expire !== Cache::PERMANENT) {
      $expire += time();
    }
    $this->cache->set($cid, $data, $expire);
  }

  /**
   * Returns API data from the persistent cache.
   *
   * @param string $cid
   *   The cache ID of the data to retrieve.
   *
   * @return mixed|false
   *   The cache data or FALSE on failure.
   */
  protected function getCacheData($cid) {
    $cache = $this->cache->get($cid);
    return $cache && isset($cache->data) ? $cache->data : FALSE;
  }

  /**
   * Calculates hash for params set.
   *
   * @param array $params
   *   A set of params to calculate a hash.
   *
   * @return string
   *   Params set hash.
   */
  protected function generateParamsHash(array $params) {
    if (empty($params)) {
      return self::PARAMS_ALL;
    }
    ksort($params);
    $params = serialize($params);
    return Crypt::hashBase64($params);
  }

  /**
   * Converts the result of XML decoding to a more convenient form.
   *
   * @param array $data
   *   An array containing the result of XML decoding.
   *
   * @return array
   *   The converted array.
   */
  protected function clearXmlDecodingResult(array $data) {
    $converted_data = [];
    foreach ($data as $key => $value) {
      // Remove unused elements.
      if ($key === '#') {
        continue;
      }
      // If the value is an array, then it must be converted.
      if (is_array($value)) {
        $value = $this->clearXmlDecodingResult($value);
      }
      // In the case of an attribute, remove the '@' character from the
      // beginning of the key.
      if (is_string($key) && strpos($key, '@') === 0) {
        $key = substr($key, 1);
      }
      $converted_data[$key] = $value;
    }
    return $converted_data;
  }

}
