<?php

/**
 * @file
 * Contains \Drupal\cdek_api\RequestManager.
 */

namespace Drupal\cdek_api;

use Psr\Http\Message\ResponseInterface;
use GuzzleHttp\Client;

/**
 * CDEK API request manager.
 */
class RequestManager {

  /**
   * Request path for order pickup points list.
   */
  const ORDER_PICKUP_POINTS_LIST_PATH = 'pvzlist.php';

  /**
   * Request path for delivery rate calculation.
   */
  const DELIVERY_RATE_CALCULATION_PATH = 'calculator/calculate_price_by_json.php';

  /**
   * CDEK API settings.
   *
   * @var array
   */
  protected $settings = [];

  /**
   * The default http client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * RequestManager constructor.
   *
   * @param \GuzzleHttp\Client $http_client
   *   A guzzle http client instance.
   */
  public function __construct(Client $http_client) {
    $this->httpClient = $http_client;
    // Get the configuration values.
    $this->settings = Cdek::getSettings();
  }

  /**
   * Returns response from "order pickup points list" API request.
   *
   * @param array $params
   *   See \Drupal\cdek_api\Cdek::getOrderPickupPoints() for details.
   *
   * @return string|null
   *   Response body or NULL on failure.
   */
  public function sendOrderPickupPointsRequest(array $params = []) {
    // Request options and URI.
    $uri = $this->settings['integration_base_path'] . '/' . self::ORDER_PICKUP_POINTS_LIST_PATH;
    $options['timeout'] = (float) $this->settings['request_timeout'];
    $options['query'] = $params;
    // Sending GET request.
    $response = $this->httpClient->get($uri, $options);
    // Check the success of the API request.
    return $this->isSuccessfulRequest($response) ? (string) $response->getBody() : NULL;
  }

  /**
   * Returns response from "calculate delivery rate" API request.
   *
   * @param array $params
   *   See \Drupal\cdek_api\Cdek::calculateDeliveryRate() for details.
   *
   * @return string|null
   *   Response body or NULL on failure.
   */
  public function sendDeliveryRateCalculationRequest(array $params = []) {
    // Request options and URI.
    $uri = $this->settings['base_path'] . '/' . self::DELIVERY_RATE_CALCULATION_PATH;
    $options['headers'] = ['Content-Type' => 'application/json'];
    $options['timeout'] = (float) $this->settings['request_timeout'];
    $options['body'] = json_encode($params);
    // Sending POST request.
    $response = $this->httpClient->post($uri, $options);
    // Check the success of the API request.
    return $this->isSuccessfulRequest($response) ? (string) $response->getBody() : NULL;
  }

  /**
   * Check the success of the API request.
   *
   * @param \Psr\Http\Message\ResponseInterface $response
   *   Response of the API request.
   *
   * @return bool
   *   The request is successful (TRUE) or not (FALSE).
   */
  protected function isSuccessfulRequest(ResponseInterface $response) {
    return $response->getStatusCode() >= 200 && $response->getStatusCode() < 300;
  }

}
